@@ -17,9 +17,11 @@
 #include <cstring>
 #include <fstream>
 #include <functional>
+#include <random>
 #include <sstream>
 #include <string>
 
+#include "source/fuzz/force_render_red.h"
 #include "source/fuzz/fuzzer.h"
 #include "source/fuzz/protobufs/spirvfuzz_protobufs.h"
 #include "source/fuzz/replayer.h"
@@ -54,6 +56,8 @@ bool ExecuteCommand(const std::string& command) {
 
 // Status and actions to perform after parsing command-line arguments.
 enum class FuzzActions {
+  FORCE_RENDER_RED,  // Turn the shader into a form such that it is guaranteed
+                     // to render a red image.
   FUZZ,    // Run the fuzzer to apply transformations in a randomized fashion.
   REPLAY,  // Replay an existing sequence of transformations.
   SHRINK,  // Shrink an existing sequence of transformations with respect to an
@@ -72,43 +76,70 @@ void PrintUsage(const char* program) {
       R"(%s - Fuzzes an equivalent SPIR-V binary based on a given binary.
 
 USAGE: %s [options] <input.spv> -o <output.spv>
+USAGE: %s [options] <input.spv> -o <output.spv> \
+  --shrink=<input.transformations> -- <interestingness_test> [args...]
 
-The SPIR-V binary is read from <input.spv>, which must have extension .spv.  If
-<input.facts> is also present, facts about the SPIR-V binary are read from this
-file.
+The SPIR-V binary is read from <input.spv>.  If <input.facts> is also present,
+facts about the SPIR-V binary are read from this file.
 
 The transformed SPIR-V binary is written to <output.spv>.  Human-readable and
 binary representations of the transformations that were applied are written to
 <output.transformations_json> and <output.transformations>, respectively.
 
+When passing --shrink=<input.transformations> an <interestingness_test>
+must also be provided; this is the path to a script that returns 0 if and only
+if a given SPIR-V binary is interesting.  The SPIR-V binary will be passed to
+the script as an argument after any other provided arguments [args...].  The
+"--" characters are optional but denote that all arguments that follow are
+positional arguments and thus will be forwarded to the interestingness script,
+and not parsed by %s.
+
 NOTE: The fuzzer is a work in progress.
 
 Options (in lexicographical order):
 
   -h, --help
                Print this help.
+  --force-render-red
+               Transforms the input shader into a shader that writes red to the
+               output buffer, and then captures the original shader as the body
+               of a conditional with a dynamically false guard.  Exploits input
+               facts to make the guard non-obviously false.  This option is a
+               helper for massaging crash-inducing tests into a runnable
+               format; it does not perform any fuzzing.
+  --fuzzer-pass-validation
+               Run the validator after applying each fuzzer pass during
+               fuzzing.  Aborts fuzzing early if an invalid binary is created.
+               Useful for debugging spirv-fuzz.
   --replay
                File from which to read a sequence of transformations to replay
                (instead of fuzzing)
-  --seed
+  --seed=
                Unsigned 32-bit integer seed to control random number
                generation.
-  --shrink
+  --shrink=
                File from which to read a sequence of transformations to shrink
                (instead of fuzzing)
-  --shrinker-step-limit
+  --shrinker-step-limit=
                Unsigned 32-bit integer specifying maximum number of steps the
                shrinker will take before giving up.  Ignored unless --shrink
                is used.
-  --interestingness
-               Path to an interestingness function to guide shrinking: a script
-               that returns 0 if and only if a given binary is interesting.
-               Required if --shrink is provided; disallowed otherwise.
+  --shrinker-temp-file-prefix=
+               Specifies a temporary file prefix that will be used to output
+               temporary shader files during shrinking.  A number and .spv
+               extension will be added.  The default is "temp_", which will
+               cause files like "temp_0001.spv" to be output to the current
+               directory.  Ignored unless --shrink is used.
+  --replay-validation
+               Run the validator after applying each transformation during
+               replay (including the replay that occurs during shrinking).
+               Aborts if an invalid binary is created.  Useful for debugging
+               spirv-fuzz.
   --version
                Display fuzzer version information.
 
 )",
-      program, program);
+      program, program, program, program);
 }
 
 // Message consumer for this tool.  Used to emit diagnostics during
@@ -122,24 +153,20 @@ void FuzzDiagnostic(spv_message_level_t level, const char* /*source*/,
   fprintf(stderr, "%s\n", message);
 }
 
-bool EndsWithSpv(const std::string& filename) {
-  std::string dot_spv = ".spv";
-  return filename.length() >= dot_spv.length() &&
-         0 == filename.compare(filename.length() - dot_spv.length(),
-                               filename.length(), dot_spv);
-}
-
 FuzzStatus ParseFlags(int argc, const char** argv, std::string* in_binary_file,
                       std::string* out_binary_file,
                       std::string* replay_transformations_file,
-                      std::string* interestingness_function_file,
+                      std::vector<std::string>* interestingness_test,
                       std::string* shrink_transformations_file,
+                      std::string* shrink_temp_file_prefix,
                       spvtools::FuzzerOptions* fuzzer_options) {
   uint32_t positional_arg_index = 0;
+  bool only_positional_arguments_remain = false;
+  bool force_render_red = false;
 
   for (int argi = 1; argi < argc; ++argi) {
     const char* cur_arg = argv[argi];
-    if ('-' == cur_arg[0]) {
+    if ('-' == cur_arg[0] && !only_positional_arguments_remain) {
       if (0 == strcmp(cur_arg, "--version")) {
         spvtools::Logf(FuzzDiagnostic, SPV_MSG_INFO, nullptr, {}, "%s\n",
                        spvSoftwareVersionDetailsString());
@@ -154,13 +181,18 @@ FuzzStatus ParseFlags(int argc, const char** argv, std::string* in_binary_file,
           PrintUsage(argv[0]);
           return {FuzzActions::STOP, 1};
         }
+      } else if (0 == strncmp(cur_arg, "--force-render-red",
+                              sizeof("--force-render-red") - 1)) {
+        force_render_red = true;
+      } else if (0 == strncmp(cur_arg, "--fuzzer-pass-validation",
+                              sizeof("--fuzzer-pass-validation") - 1)) {
+        fuzzer_options->enable_fuzzer_pass_validation();
       } else if (0 == strncmp(cur_arg, "--replay=", sizeof("--replay=") - 1)) {
         const auto split_flag = spvtools::utils::SplitFlagArgs(cur_arg);
         *replay_transformations_file = std::string(split_flag.second);
-      } else if (0 == strncmp(cur_arg, "--interestingness=",
-                              sizeof("--interestingness=") - 1)) {
-        const auto split_flag = spvtools::utils::SplitFlagArgs(cur_arg);
-        *interestingness_function_file = std::string(split_flag.second);
+      } else if (0 == strncmp(cur_arg, "--replay-validation",
+                              sizeof("--replay-validation") - 1)) {
+        fuzzer_options->enable_replay_validation();
       } else if (0 == strncmp(cur_arg, "--shrink=", sizeof("--shrink=") - 1)) {
         const auto split_flag = spvtools::utils::SplitFlagArgs(cur_arg);
         *shrink_transformations_file = std::string(split_flag.second);
@@ -181,11 +213,18 @@ FuzzStatus ParseFlags(int argc, const char** argv, std::string* in_binary_file,
             static_cast<uint32_t>(strtol(split_flag.second.c_str(), &end, 10));
         assert(end != split_flag.second.c_str() && errno == 0);
         fuzzer_options->set_shrinker_step_limit(step_limit);
-      } else if ('\0' == cur_arg[1]) {
-        // We do not support fuzzing from standard input.  We could support
-        // this if there was a compelling use case.
+      } else if (0 == strncmp(cur_arg, "--shrinker-temp-file-prefix=",
+                              sizeof("--shrinker-temp-file-prefix=") - 1)) {
+        const auto split_flag = spvtools::utils::SplitFlagArgs(cur_arg);
+        *shrink_temp_file_prefix = std::string(split_flag.second);
+      } else if (0 == strcmp(cur_arg, "--")) {
+        only_positional_arguments_remain = true;
+      } else {
+        std::stringstream ss;
+        ss << "Unrecognized argument: " << cur_arg << std::endl;
+        spvtools::Error(FuzzDiagnostic, nullptr, {}, ss.str().c_str());
         PrintUsage(argv[0]);
-        return {FuzzActions::STOP, 0};
+        return {FuzzActions::STOP, 1};
       }
     } else if (positional_arg_index == 0) {
       // Binary input file name
@@ -193,9 +232,7 @@ FuzzStatus ParseFlags(int argc, const char** argv, std::string* in_binary_file,
       *in_binary_file = std::string(cur_arg);
       positional_arg_index++;
     } else {
-      spvtools::Error(FuzzDiagnostic, nullptr, {},
-                      "Too many positional arguments specified");
-      return {FuzzActions::STOP, 1};
+      interestingness_test->push_back(std::string(cur_arg));
     }
   }
 
@@ -204,20 +241,47 @@ FuzzStatus ParseFlags(int argc, const char** argv, std::string* in_binary_file,
     return {FuzzActions::STOP, 1};
   }
 
-  if (!EndsWithSpv(*in_binary_file)) {
-    spvtools::Error(FuzzDiagnostic, nullptr, {},
-                    "Input filename must have extension .spv");
-    return {FuzzActions::STOP, 1};
-  }
-
   if (out_binary_file->empty()) {
     spvtools::Error(FuzzDiagnostic, nullptr, {}, "-o required");
     return {FuzzActions::STOP, 1};
   }
 
-  if (!EndsWithSpv(*out_binary_file)) {
+  auto const_fuzzer_options =
+      static_cast<spv_const_fuzzer_options>(*fuzzer_options);
+  if (force_render_red) {
+    if (!replay_transformations_file->empty() ||
+        !shrink_transformations_file->empty() ||
+        const_fuzzer_options->replay_validation_enabled) {
+      spvtools::Error(FuzzDiagnostic, nullptr, {},
+                      "The --force-render-red argument cannot be used with any "
+                      "other arguments except -o.");
+      return {FuzzActions::STOP, 1};
+    }
+    return {FuzzActions::FORCE_RENDER_RED, 0};
+  }
+
+  if (replay_transformations_file->empty() &&
+      shrink_transformations_file->empty() &&
+      static_cast<spv_const_fuzzer_options>(*fuzzer_options)
+          ->replay_validation_enabled) {
     spvtools::Error(FuzzDiagnostic, nullptr, {},
-                    "Output filename must have extension .spv");
+                    "The --replay-validation argument can only be used with "
+                    "one of the --replay or --shrink arguments.");
+    return {FuzzActions::STOP, 1};
+  }
+
+  if (shrink_transformations_file->empty() && !interestingness_test->empty()) {
+    spvtools::Error(FuzzDiagnostic, nullptr, {},
+                    "Too many positional arguments specified; extra positional "
+                    "arguments are used as the interestingness function, which "
+                    "are only valid with the --shrink option.");
+    return {FuzzActions::STOP, 1};
+  }
+
+  if (!shrink_transformations_file->empty() && interestingness_test->empty()) {
+    spvtools::Error(
+        FuzzDiagnostic, nullptr, {},
+        "The --shrink option requires an interestingness function.");
     return {FuzzActions::STOP, 1};
   }
 
@@ -230,28 +294,14 @@ FuzzStatus ParseFlags(int argc, const char** argv, std::string* in_binary_file,
           "The --replay and --shrink arguments are mutually exclusive.");
       return {FuzzActions::STOP, 1};
     }
-    if (!interestingness_function_file->empty()) {
-      spvtools::Error(FuzzDiagnostic, nullptr, {},
-                      "The --replay and --interestingness arguments are "
-                      "mutually exclusive.");
-      return {FuzzActions::STOP, 1};
-    }
     return {FuzzActions::REPLAY, 0};
   }
 
-  if (!shrink_transformations_file->empty() ^
-      !interestingness_function_file->empty()) {
-    spvtools::Error(FuzzDiagnostic, nullptr, {},
-                    "Both or neither of the --shrink and --interestingness "
-                    "arguments must be provided.");
-    return {FuzzActions::STOP, 1};
-  }
-
   if (!shrink_transformations_file->empty()) {
     // The tool is being invoked in shrink mode.
-    assert(!interestingness_function_file->empty() &&
-           "An error should have been raised if --shrink but not --interesting "
-           "was provided.");
+    assert(!interestingness_test->empty() &&
+           "An error should have been raised if --shrink was provided without "
+           "an interestingness test.");
     return {FuzzActions::SHRINK, 0};
   }
 
@@ -278,6 +328,7 @@ bool ParseTransformations(
 }
 
 bool Replay(const spv_target_env& target_env,
+            spv_const_fuzzer_options fuzzer_options,
             const std::vector<uint32_t>& binary_in,
             const spvtools::fuzz::protobufs::FactSequence& initial_facts,
             const std::string& replay_transformations_file,
@@ -289,7 +340,8 @@ bool Replay(const spv_target_env& target_env,
                             &transformation_sequence)) {
     return false;
   }
-  spvtools::fuzz::Replayer replayer(target_env);
+  spvtools::fuzz::Replayer replayer(target_env,
+                                    fuzzer_options->replay_validation_enabled);
   replayer.SetMessageConsumer(spvtools::utils::CLIMessageConsumer);
   auto replay_result_status =
       replayer.Run(binary_in, initial_facts, transformation_sequence,
@@ -303,7 +355,8 @@ bool Shrink(const spv_target_env& target_env,
             const std::vector<uint32_t>& binary_in,
             const spvtools::fuzz::protobufs::FactSequence& initial_facts,
             const std::string& shrink_transformations_file,
-            const std::string& interestingness_function_file,
+            const std::string& shrink_temp_file_prefix,
+            const std::vector<std::string>& interestingness_command,
             std::vector<uint32_t>* binary_out,
             spvtools::fuzz::protobufs::TransformationSequence*
                 transformations_applied) {
@@ -313,18 +366,28 @@ bool Shrink(const spv_target_env& target_env,
     return false;
   }
   spvtools::fuzz::Shrinker shrinker(target_env,
-                                    fuzzer_options->shrinker_step_limit);
+                                    fuzzer_options->shrinker_step_limit,
+                                    fuzzer_options->replay_validation_enabled);
   shrinker.SetMessageConsumer(spvtools::utils::CLIMessageConsumer);
 
+  assert(!interestingness_command.empty() &&
+         "An error should have been raised because the interestingness_command "
+         "is empty.");
+  std::stringstream joined;
+  joined << interestingness_command[0];
+  for (size_t i = 1, size = interestingness_command.size(); i < size; ++i) {
+    joined << " " << interestingness_command[i];
+  }
+  std::string interestingness_command_joined = joined.str();
+
   spvtools::fuzz::Shrinker::InterestingnessFunction interestingness_function =
-      [interestingness_function_file](std::vector<uint32_t> binary,
-                                      uint32_t reductions_applied) -> bool {
+      [interestingness_command_joined, shrink_temp_file_prefix](
+          std::vector<uint32_t> binary, uint32_t reductions_applied) -> bool {
     std::stringstream ss;
-    ss << "temp_" << std::setw(4) << std::setfill('0') << reductions_applied
-       << ".spv";
+    ss << shrink_temp_file_prefix << std::setw(4) << std::setfill('0')
+       << reductions_applied << ".spv";
     const auto spv_file = ss.str();
-    const std::string command =
-        std::string(interestingness_function_file) + " " + spv_file;
+    const std::string command = interestingness_command_joined + " " + spv_file;
     auto write_file_succeeded =
         WriteFile(spv_file.c_str(), "wb", &binary[0], binary.size());
     (void)(write_file_succeeded);
@@ -342,16 +405,21 @@ bool Shrink(const spv_target_env& target_env,
 }
 
 bool Fuzz(const spv_target_env& target_env,
-          const spvtools::FuzzerOptions& fuzzer_options,
+          spv_const_fuzzer_options fuzzer_options,
           const std::vector<uint32_t>& binary_in,
           const spvtools::fuzz::protobufs::FactSequence& initial_facts,
           std::vector<uint32_t>* binary_out,
           spvtools::fuzz::protobufs::TransformationSequence*
               transformations_applied) {
-  spvtools::fuzz::Fuzzer fuzzer(target_env);
+  spvtools::fuzz::Fuzzer fuzzer(
+      target_env,
+      fuzzer_options->has_random_seed
+          ? fuzzer_options->random_seed
+          : static_cast<uint32_t>(std::random_device()()),
+      fuzzer_options->fuzzer_pass_validation_enabled);
   fuzzer.SetMessageConsumer(spvtools::utils::CLIMessageConsumer);
-  auto fuzz_result_status = fuzzer.Run(binary_in, initial_facts, fuzzer_options,
-                                       binary_out, transformations_applied);
+  auto fuzz_result_status =
+      fuzzer.Run(binary_in, initial_facts, binary_out, transformations_applied);
   if (fuzz_result_status !=
       spvtools::fuzz::Fuzzer::FuzzerResultStatus::kComplete) {
     spvtools::Error(FuzzDiagnostic, nullptr, {}, "Error running fuzzer");
@@ -362,21 +430,39 @@ bool Fuzz(const spv_target_env& target_env,
 
 }  // namespace
 
+// Dumps |binary| to file |filename|. Useful for interactive debugging.
+void DumpShader(const std::vector<uint32_t>& binary, const char* filename) {
+  auto write_file_succeeded =
+      WriteFile(filename, "wb", &binary[0], binary.size());
+  if (!write_file_succeeded) {
+    std::cerr << "Failed to dump shader" << std::endl;
+  }
+}
+
+// Dumps the SPIRV-V module in |context| to file |filename|. Useful for
+// interactive debugging.
+void DumpShader(spvtools::opt::IRContext* context, const char* filename) {
+  std::vector<uint32_t> binary;
+  context->module()->ToBinary(&binary, false);
+  DumpShader(binary, filename);
+}
+
 const auto kDefaultEnvironment = SPV_ENV_UNIVERSAL_1_3;
 
 int main(int argc, const char** argv) {
   std::string in_binary_file;
   std::string out_binary_file;
   std::string replay_transformations_file;
-  std::string interestingness_function_file;
+  std::vector<std::string> interestingness_test;
   std::string shrink_transformations_file;
+  std::string shrink_temp_file_prefix = "temp_";
 
   spvtools::FuzzerOptions fuzzer_options;
 
-  FuzzStatus status =
-      ParseFlags(argc, argv, &in_binary_file, &out_binary_file,
-                 &replay_transformations_file, &interestingness_function_file,
-                 &shrink_transformations_file, &fuzzer_options);
+  FuzzStatus status = ParseFlags(
+      argc, argv, &in_binary_file, &out_binary_file,
+      &replay_transformations_file, &interestingness_test,
+      &shrink_transformations_file, &shrink_temp_file_prefix, &fuzzer_options);
 
   if (status.action == FuzzActions::STOP) {
     return status.code;
@@ -388,10 +474,12 @@ int main(int argc, const char** argv) {
   }
 
   spvtools::fuzz::protobufs::FactSequence initial_facts;
-  const std::string dot_spv(".spv");
-  std::string in_facts_file =
-      in_binary_file.substr(0, in_binary_file.length() - dot_spv.length()) +
-      ".facts";
+
+  // If not found, dot_pos will be std::string::npos, which can be used in
+  // substr to mean "the end of the string"; there is no need to check the
+  // result.
+  size_t dot_pos = in_binary_file.rfind('.');
+  std::string in_facts_file = in_binary_file.substr(0, dot_pos) + ".facts";
   std::ifstream facts_input(in_facts_file);
   if (facts_input) {
     std::string facts_json_string((std::istreambuf_iterator<char>(facts_input)),
@@ -411,6 +499,12 @@ int main(int argc, const char** argv) {
   spv_target_env target_env = kDefaultEnvironment;
 
   switch (status.action) {
+    case FuzzActions::FORCE_RENDER_RED:
+      if (!spvtools::fuzz::ForceRenderRed(target_env, binary_in, initial_facts,
+                                          &binary_out)) {
+        return 1;
+      }
+      break;
     case FuzzActions::FUZZ:
       if (!Fuzz(target_env, fuzzer_options, binary_in, initial_facts,
                 &binary_out, &transformations_applied)) {
@@ -418,7 +512,7 @@ int main(int argc, const char** argv) {
       }
       break;
     case FuzzActions::REPLAY:
-      if (!Replay(target_env, binary_in, initial_facts,
+      if (!Replay(target_env, fuzzer_options, binary_in, initial_facts,
                   replay_transformations_file, &binary_out,
                   &transformations_applied)) {
         return 1;
@@ -431,8 +525,9 @@ int main(int argc, const char** argv) {
         return 1;
       }
       if (!Shrink(target_env, fuzzer_options, binary_in, initial_facts,
-                  shrink_transformations_file, interestingness_function_file,
-                  &binary_out, &transformations_applied)) {
+                  shrink_transformations_file, shrink_temp_file_prefix,
+                  interestingness_test, &binary_out,
+                  &transformations_applied)) {
         return 1;
       }
     } break;
@@ -447,35 +542,40 @@ int main(int argc, const char** argv) {
     return 1;
   }
 
-  std::string output_file_prefix =
-      out_binary_file.substr(0, out_binary_file.length() - dot_spv.length());
-  std::ofstream transformations_file;
-  transformations_file.open(output_file_prefix + ".transformations",
-                            std::ios::out | std::ios::binary);
-  bool success =
-      transformations_applied.SerializeToOstream(&transformations_file);
-  transformations_file.close();
-  if (!success) {
-    spvtools::Error(FuzzDiagnostic, nullptr, {},
-                    "Error writing out transformations binary");
-    return 1;
-  }
+  if (status.action != FuzzActions::FORCE_RENDER_RED) {
+    // If not found, dot_pos will be std::string::npos, which can be used in
+    // substr to mean "the end of the string"; there is no need to check the
+    // result.
+    dot_pos = out_binary_file.rfind('.');
+    std::string output_file_prefix = out_binary_file.substr(0, dot_pos);
+    std::ofstream transformations_file;
+    transformations_file.open(output_file_prefix + ".transformations",
+                              std::ios::out | std::ios::binary);
+    bool success =
+        transformations_applied.SerializeToOstream(&transformations_file);
+    transformations_file.close();
+    if (!success) {
+      spvtools::Error(FuzzDiagnostic, nullptr, {},
+                      "Error writing out transformations binary");
+      return 1;
+    }
 
-  std::string json_string;
-  auto json_options = google::protobuf::util::JsonOptions();
-  json_options.add_whitespace = true;
-  auto json_generation_status = google::protobuf::util::MessageToJsonString(
-      transformations_applied, &json_string, json_options);
-  if (json_generation_status != google::protobuf::util::Status::OK) {
-    spvtools::Error(FuzzDiagnostic, nullptr, {},
-                    "Error writing out transformations in JSON format");
-    return 1;
-  }
+    std::string json_string;
+    auto json_options = google::protobuf::util::JsonOptions();
+    json_options.add_whitespace = true;
+    auto json_generation_status = google::protobuf::util::MessageToJsonString(
+        transformations_applied, &json_string, json_options);
+    if (json_generation_status != google::protobuf::util::Status::OK) {
+      spvtools::Error(FuzzDiagnostic, nullptr, {},
+                      "Error writing out transformations in JSON format");
+      return 1;
+    }
 
-  std::ofstream transformations_json_file(output_file_prefix +
-                                          ".transformations_json");
-  transformations_json_file << json_string;
-  transformations_json_file.close();
+    std::ofstream transformations_json_file(output_file_prefix +
+                                            ".transformations_json");
+    transformations_json_file << json_string;
+    transformations_json_file.close();
+  }
 
   return 0;
 }
